/*
  A simple ASN.1 parser, similiar to 'dumpasn1' or 'openssl asn1parse', though
  without some of the bells and whistles of those. Primarily used for testing
  the BER decoder. The output format is modeled loosely on 'asn1parse -i'

  The output is actually less precise than the other decoders named, because
  the underlying BER_Decoder hides quite a bit from userspace, such as the use
  of indefinite length encodings (and the EOC markers). At some point it will
  also hide the constructed string types from the user, but right now you'll
  seem them as-is.

  Written by Jack Lloyd, November 9-10, 2003

  This file is in the public domain.
*/

#include <botan/ber_dec.h>
#include <botan/asn1_obj.h>
#include <botan/pem.h>
#include <botan/oids.h>
#include <botan/pipe.h>
#include <botan/hex.h>
#include <botan/pem.h>
using namespace Botan;

/* Technically bad C++ style, but printf is just so much easier to use than
   cout for things like this...
*/
#include <stdio.h>
#include <ctype.h>

void decode(BER_Decoder&, u32bit);
void emit(const std::string&, u32bit, u32bit, const std::string& = "");
std::string type_name(ASN1_Tag);

/*
   What level the outermost layer of stuff is at. Probably 0 or 1; asn1parse
   uses 0 as the outermost, while 1 makes more sense to me. 2+ doesn't make
   much sense at all.
*/
const u32bit INITIAL_LEVEL = 0;

int main(int argc, char* argv[])
   {
   if(argc != 2)
      {
      printf("Usage: %s <file>\n", argv[0]);
      return 1;
      }

   try {
      LibraryInitializer init;

      DataSource_Stream in(argv[1]);

      if(BER::maybe_BER(in) && !PEM_Code::matches(in))
         {
         BER_Decoder decoder(in);
         decode(decoder, INITIAL_LEVEL);
         }
      else
         {
         std::string label; // ignored
         BER_Decoder decoder(PEM_Code::decode(in, label));
         decode(decoder, INITIAL_LEVEL);
         }

   }
   catch(std::exception& e)
      {
      printf("%s\n", e.what());
      return 1;
      }
   return 0;
   }

void decode(BER_Decoder& decoder, u32bit level)
   {
   BER_Object obj = decoder.get_next_object();
   while(obj.tag != NO_OBJECT)
      {
      const ASN1_Tag tag = obj.tag;
      const u32bit length = obj.value.size();

      /* hack to insert the tag+length back in front of the stuff now
         that we've gotten the type info */
      DER_Encoder encoder;
      encoder.add_object(tag, UNIVERSAL, obj.value, obj.value.size());
      SecureVector<byte> bits = encoder.get_contents();

      BER_Decoder data(bits);

      if(tag & CONSTRUCTED)
         {
         ASN1_Tag con_type = ASN1_Tag(tag & 0x1F);

         BER_Decoder cons_info(obj.value);
         if(con_type == SEQUENCE)
            {
            emit("SEQUENCE", level, length);
            decode(cons_info, level+1);
            }
         else if(con_type == SET)
            {
            emit("SET", level, length);
            decode(cons_info, level+1);
            }
         else
            {
            std::string name;

            if((tag & APPLICATION) || (tag & CONTEXT_SPECIFIC) ||
               (tag & PRIVATE))
               name = "cons [" + to_string(con_type) + "]";
            else
               name = type_name(con_type) + " (cons)";

            emit(name, level, length);
            decode(cons_info, level+1);
            }
         }
      else if(tag == OBJECT_ID)
         {
         OID oid;
         BER::decode(data, oid);
         emit(type_name(tag), level, length, OIDS::lookup(oid));
         }
      else if(tag == INTEGER)
         {
         BigInt number;
         BER::decode(data, number);

         SecureVector<byte> rep;

         /* If it's small, it's probably a real number, not a hash */
         if(number.bits() <= 16)
            rep = BigInt::encode(number, BigInt::Decimal);
         else
            rep = BigInt::encode(number, BigInt::Hexadecimal);

         std::string str;
         for(u32bit j = 0; j != rep.size(); j++)
            str += (char)rep[j];

         emit(type_name(tag), level, length, str);
         }
      else if(tag == BOOLEAN)
         {
         bool boolean;
         BER::decode(data, boolean);
         emit(type_name(tag), level, length, (boolean ? "true" : "false"));
         }
      else if(tag == NULL_TAG)
         {
         emit(type_name(tag), level, length);
         }
      else if(tag == OCTET_STRING || tag == BIT_STRING)
         {
         SecureVector<byte> bits;
         BER::decode(data, bits, tag);
         bool not_text = false;

         for(u32bit j = 0; j != bits.size(); j++)
            if(!isgraph(bits[j]) && !isspace(bits[j]))
               not_text = true;

         Filter* encoder = 0;

         if(tag == BIT_STRING || not_text)
            encoder = new Hex_Encoder;

         Pipe pipe(encoder);
         pipe.process_msg(bits);

         emit(type_name(tag), level, length, pipe.read_all_as_string());
         }
      else if(tag == PRINTABLE_STRING || tag == NUMERIC_STRING ||
              tag == IA5_STRING || tag == T61_STRING)
         {
         ASN1_String str;
         BER::decode(data, str);
         emit(type_name(tag), level, length, str.value());
         }
      else if(tag == UTC_TIME || tag == GENERALIZED_TIME)
         {
         X509_Time time;
         BER::decode(data, time);
         emit(type_name(tag), level, length, time.readable_string());
         }
      else
         fprintf(stderr, "Unknown tag %02X\n", tag);

      obj = decoder.get_next_object();
      }
   }

void emit(const std::string& type, u32bit level, u32bit length,
          const std::string& value)
   {
   const u32bit LIMIT = 128;
   const u32bit BIN_LIMIT = 64;

   int written = 0;
   written += printf("  d=%2d, l=%4d: ", level, length);
   for(u32bit j = INITIAL_LEVEL; j != level; j++)
      written += printf(" ");
   written += printf("%s   ", type.c_str());

   bool should_skip = false;
   if(value.length() > LIMIT) should_skip = true;
   if((type == "OCTET STRING" || type == "BIT STRING") &&
      value.length() > BIN_LIMIT)
      should_skip = true;

   if(value != "" && !should_skip)
      {
      if(written % 2 == 0) printf(" ");
      while(written < 50) written += printf("  ");
      printf(":%s\n", value.c_str());
      }
   else
      printf("\n");
   }

std::string type_name(ASN1_Tag type)
   {
   if(type == PRINTABLE_STRING) return "PRINTABLE STRING";
   if(type == NUMERIC_STRING)   return "NUMERIC STRING";
   if(type == IA5_STRING)       return "IA5 STRING";
   if(type == T61_STRING)       return "T61 STRING";
   if(type == UTF8_STRING)      return "UTF8 STRING";
   if(type == VISIBLE_STRING)   return "VISIBLE STRING";
   if(type == BMP_STRING)       return "BMP STRING";

   if(type == UTC_TIME)         return "UTC TIME";
   if(type == GENERALIZED_TIME) return "GENERALIZED TIME";

   if(type == OCTET_STRING)     return "OCTET STRING";
   if(type == BIT_STRING)       return "BIT STRING";

   if(type == INTEGER)          return "INTEGER";
   if(type == NULL_TAG)         return "NULL";
   if(type == OBJECT_ID)        return "OBJECT";
   return "(UNKNOWN)";
   }
